/***************************************************************************
 $RCSfile: kbanking.cpp,v $
                             -------------------
    cvs         : $Id: kbanking.cpp,v 1.27 2005/03/30 08:26:20 aquamaniac Exp $
    begin       : Mon Mar 01 2004
    copyright   : (C) 2004 by Martin Preuss
    email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "kbanking_p.h"
#include "progress.h"
#include "simplebox.h"
#include "inputbox.h"
#include "ktimport.h"
#include "flagstaff.h"
#include "printdialog.h"
#include "checkduplicates.h"
#include "transactionwindow.h"
#include "mainwindow.h"

#include "report.h"
#include "simplereport.h"

#include <assert.h>
#include <qstring.h>
#include <qmessagebox.h>
#include <qinputdialog.h>
#include <qapplication.h>
#include <qdatetime.h>

#include <gwenhywfar/debug.h>
#include <gwenhywfar/text.h>

#include <qwidget.h>


KBanking::KBanking(const char *appname,
                   const char *fname,
                   const char *name)
:App(appname, fname)
,_mainWindow(0)
,_lastWidgetId(0)
,_logLevel(AB_Banking_LogLevelInfo)
,_lastReportId(0){
  _flagStaff=new FlagStaff();
}



KBanking::~KBanking(){
  delete _flagStaff;
}



int KBanking::_extractHTML(const char *text, GWEN_BUFFER *tbuf) {
  GWEN_BUFFEREDIO *bio;
  GWEN_XMLNODE *xmlNode;
  GWEN_BUFFER *buf;
  int rv;

  buf=GWEN_Buffer_new(0, 256, 0, 1);
  GWEN_Buffer_AppendString(buf, text);
  GWEN_Buffer_Rewind(buf);
  bio=GWEN_BufferedIO_Buffer2_new(buf, 1);
  GWEN_BufferedIO_SetReadBuffer(bio, 0, 256);
  xmlNode=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "html");
  rv=GWEN_XML_Parse(xmlNode, bio,
		    GWEN_XML_FLAGS_DEFAULT |
                    GWEN_XML_FLAGS_HANDLE_OPEN_HTMLTAGS);
  GWEN_BufferedIO_Close(bio);
  GWEN_BufferedIO_free(bio);
  if (rv) {
    DBG_INFO(0, "No or bad HTML");
    GWEN_XMLNode_free(xmlNode);
    return -1;
  }
  else {
    GWEN_XMLNODE *nn;

    nn=GWEN_XMLNode_FindFirstTag(xmlNode, "html", 0, 0);
    if (nn) {
      GWEN_XMLNODE *on, *onn;

      on=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "root");
      onn=GWEN_XMLNode_new(GWEN_XMLNodeTypeTag, "qt");
      GWEN_XMLNode_AddChild(on, onn);
      GWEN_XMLNode_AddChildrenOnly(onn, nn, 1);

        /* text contains HTML tag, take it */
      bio=GWEN_BufferedIO_Buffer2_new(tbuf, 0);
      GWEN_BufferedIO_SetWriteBuffer(bio, 0, 256);
      rv=GWEN_XMLNode_WriteToStream(on, bio, GWEN_XML_FLAGS_DEFAULT);
      GWEN_XMLNode_free(on);
      if (rv) {
	DBG_ERROR(AQBANKING_LOGDOMAIN, "Error writing data to stream");
	GWEN_BufferedIO_Abandon(bio);
	GWEN_BufferedIO_free(bio);
        GWEN_XMLNode_free(xmlNode);
	return -1;
      }
      else {
	rv=GWEN_BufferedIO_Close(bio);
	GWEN_BufferedIO_free(bio);
        if (rv) {
          GWEN_XMLNode_free(xmlNode);
          return -1;
        }
      }
    }
    else {
      GWEN_XMLNode_free(xmlNode);
      return 1;
    }
  }
  GWEN_XMLNode_free(xmlNode);
  return 0;
}



std::string KBanking::guiString(const char *s) {
  GWEN_BUFFER *tbuf;
  std::string res;

  assert(s);
  tbuf=GWEN_Buffer_new(0, strlen(s), 0, 1);
  if (_extractHTML(s, tbuf)) {
    GWEN_Buffer_free(tbuf);
    return s;
  }
  res=std::string(GWEN_Buffer_GetStart(tbuf),
                  GWEN_Buffer_GetUsedBytes(tbuf));
  GWEN_Buffer_free(tbuf);
  return res;
}



int KBanking::messageBox(GWEN_TYPE_UINT32 flags,
                         const char *title,
                         const char *text,
                         const char *b1,
                         const char *b2,
                         const char *b3){
  int rv;
  GWEN_BUFFER *buf;

  buf=GWEN_Buffer_new(0, strlen(text), 0, 1);
  if (!_extractHTML(text, buf)) {
    text=GWEN_Buffer_GetStart(buf);
  }

  switch(flags & AB_BANKING_MSG_FLAGS_TYPE_MASK) {
  case AB_BANKING_MSG_FLAGS_TYPE_WARN:
    rv=QMessageBox::warning(0, QString::fromUtf8(title), QString::fromUtf8(text), 
			    b1 ? QString::fromUtf8(b1) : QString::null,
			    b2 ? QString::fromUtf8(b2) : QString::null,
			    b3 ? QString::fromUtf8(b3) : QString::null);
    break;
  case AB_BANKING_MSG_FLAGS_TYPE_ERROR:
    rv=QMessageBox::critical(0, QString::fromUtf8(title), QString::fromUtf8(text), 
			     b1 ? QString::fromUtf8(b1) : QString::null,
			     b2 ? QString::fromUtf8(b2) : QString::null,
			     b3 ? QString::fromUtf8(b3) : QString::null);
    break;
  case AB_BANKING_MSG_FLAGS_TYPE_INFO:
  default:
    rv=QMessageBox::information(0, QString::fromUtf8(title), QString::fromUtf8(text),
				b1 ? QString::fromUtf8(b1) : QString::null,
				b2 ? QString::fromUtf8(b2) : QString::null,
				b3 ? QString::fromUtf8(b3) : QString::null);
    break;
  }
  rv++;
  GWEN_Buffer_free(buf);
  return rv;
}



int KBanking::inputBox(GWEN_TYPE_UINT32 flags,
                       const char *title,
                       const char *text,
                       char *buffer,
                       int minLen,
                       int maxLen){
  GWEN_BUFFER *buf;

  buf=GWEN_Buffer_new(0, strlen(text), 0, 1);
  if (!_extractHTML(text, buf)) {
    text=GWEN_Buffer_GetStart(buf);
  }

  KBInputBox ib(QString::fromUtf8(title), QString::fromUtf8(text),
		flags, minLen, maxLen, 0, "InputBox", true);
  GWEN_Buffer_free(buf);
  if (ib.exec()==QDialog::Accepted) {
    QString s;
    int l;

    s=ib.getInput();
    l=s.length();
    if (l && l<maxLen-1) {
      memmove(buffer, s.latin1(), l);
      buffer[l]=0;
    }
    else {
      DBG_ERROR(0, "Bad pin length");
      return AB_ERROR_INVALID;
    }
    return 0;
  }
  else {
    DBG_WARN(0, "Aborted by user");
    return AB_ERROR_USER_ABORT;
  }
}



GWEN_TYPE_UINT32 KBanking::showBox(GWEN_TYPE_UINT32 flags,
                                   const char *title,
                                   const char *text){
  GWEN_TYPE_UINT32 id;
  KBSimpleBox *b;
  QWidget *w;
  GWEN_BUFFER *buf;

  buf=GWEN_Buffer_new(0, strlen(text), 0, 1);
  if (!_extractHTML(text, buf)) {
    text=GWEN_Buffer_GetStart(buf);
  }

  w=_findProgressWidget(0);
  id=++_lastWidgetId;
  b=new KBSimpleBox(id, QString::fromUtf8(title),
		    QString::fromUtf8(text),
		    w, "SimpleBox",
		    Qt::WType_TopLevel | Qt::WType_Dialog | Qt::WShowModal);
  GWEN_Buffer_free(buf);
  //b->setModal(true);
  if (flags & AB_BANKING_SHOWBOX_FLAGS_BEEP)
    QApplication::beep();

  b->show();
  _simpleBoxWidgets.push_front(b);
  qApp->processEvents();
  return id;
}


void KBanking::hideBox(GWEN_TYPE_UINT32 id){
  if (_simpleBoxWidgets.size()==0) {
    DBG_WARN(0, "No simpleBox widgets");
    return;
  }
  if (id==0) {
    KBSimpleBox *b;

    b=_simpleBoxWidgets.front();
    b->close(true);
    _simpleBoxWidgets.pop_front();
  }
  else {
    std::list<KBSimpleBox*>::iterator it;
    for (it=_simpleBoxWidgets.begin(); it!=_simpleBoxWidgets.end(); it++) {
      if ((*it)->getId()==id) {
        (*it)->close(true);
        _simpleBoxWidgets.erase(it);
        break;
      }
    }
  }
  qApp->processEvents();
}



GWEN_TYPE_UINT32 KBanking::progressStart(const char *title,
                                         const char *text,
                                         GWEN_TYPE_UINT32 total){
  GWEN_TYPE_UINT32 id;
  KBProgress *pr;
  GWEN_BUFFER *buf;

  _cleanupProgressWidgets();
  id=++_lastWidgetId;
  buf=GWEN_Buffer_new(0, strlen(text)+1, 0, 1);
  if (!_extractHTML(text, buf)) {
    text=GWEN_Buffer_GetStart(buf);
  }

  pr=new KBProgress(id, QString::fromUtf8(title), 
                    QString::fromUtf8(text),
                    _mainWindow, "ProgressWidget",
                    Qt::WType_Dialog | Qt::WShowModal);
  GWEN_Buffer_free(buf);
  if (pr->start(total)) {
    DBG_ERROR(0, "Could not start progress dialog");
    delete pr;
    return 0;
  }
  pr->show();
  _progressWidgets.push_front(pr);
  return id;
}



KBProgress *KBanking::_findProgressWidget(GWEN_TYPE_UINT32 id) {
  std::list<KBProgress*>::iterator it;

  if (_progressWidgets.size()==0) {
    DBG_INFO(0, "No progress widgets");
    return 0;
  }
  if (id==0)
    return _progressWidgets.front();
  for (it=_progressWidgets.begin(); it!=_progressWidgets.end(); it++) {
    if ((*it)->getId()==id)
      return *it;
  }

  DBG_WARN(0, "Progress widget \"%08x\" not found", (unsigned int)id);
  return 0;
}



void KBanking::_cleanupProgressWidgets() {
  std::list<KBProgress*>::iterator it;

  while(1) {
    bool del;

    del=false;
    for (it=_progressWidgets.begin(); it!=_progressWidgets.end(); it++) {
      if ((*it)->isClosed()) {
        delete *it;
        _progressWidgets.erase(it);
        del=true;
        break;
      }
    } /* for */
    if (!del)
      break;
  } /* while */
}



int KBanking::progressAdvance(GWEN_TYPE_UINT32 id,
                              GWEN_TYPE_UINT32 progress){
  KBProgress *pr;

  pr=_findProgressWidget(id);
  if (pr) {
    return pr->advance(progress);
  }
  else {
    return AB_ERROR_NOT_FOUND;
  }
}



int KBanking::progressLog(GWEN_TYPE_UINT32 id,
                          AB_BANKING_LOGLEVEL level,
                          const char *chartext){
  KBProgress *pr;
  QString text(QString::fromUtf8(chartext));

  // Necessary when passing this QString into the macros
  const char *latin1text = text.latin1();

  if (level>_logLevel) {
    DBG_NOTICE(0, "Not logging this: %02d: %s (we are at %d)",
               level, latin1text, _logLevel);
    /* don't log this */
    return 0;
  }

  DBG_INFO(0, "%02d: %s", level, latin1text);
  pr=_findProgressWidget(id);
  if (pr) {
    return pr->log(level, text);
  }
  else {
    DBG_ERROR(0, "No progress widget currently open");
    return AB_ERROR_NOT_FOUND;
  }
}



int KBanking::progressEnd(GWEN_TYPE_UINT32 id){
  KBProgress *pr;
  int res;

  if (_progressWidgets.size()==0) {
    DBG_INFO(0, "No active progress widget");
    return AB_ERROR_NOT_FOUND;
  }

  res=AB_ERROR_NOT_FOUND;
  if (id==0) {
    pr=_progressWidgets.front();
    res=pr->end();
    _progressWidgets.pop_front();
  }
  else {
    std::list<KBProgress*>::iterator it;

    for (it=_progressWidgets.begin(); it!=_progressWidgets.end(); it++) {
      if ((*it)->getId()==id) {
	res=(*it)->end();
	_progressWidgets.erase(it);
	break;
      }
    }
  }

  return res;
}



FlagStaff *KBanking::flagStaff(){
  return _flagStaff;
}



int KBanking::enqueueJob(AB_JOB *j){
  int rv;

  rv=App::enqueueJob(j);

  _flagStaff->queueUpdated();
  return rv;
}



int KBanking::dequeueJob(AB_JOB *j){
  int rv;

  rv=App::Banking::dequeueJob(j);
  _flagStaff->queueUpdated();
  return rv;
}



int KBanking::executeQueue(){
  int rv;

  rv=App::Banking::executeQueue();
  _flagStaff->queueUpdated();
  return rv;
}



void KBanking::accountsUpdated(){
  _flagStaff->accountsUpdated();
}



void KBanking::transfersUpdated(){
  _flagStaff->transfersUpdated();
}



bool KBanking::importContext(AB_IMEXPORTER_CONTEXT *ctx,
                             GWEN_TYPE_UINT32 flags){
  KTransactionImporter imp(this, ctx, flags);

  DBG_INFO(0, "Checking context...");
  if (!imp.checkTransactions()) {
    DBG_INFO(0, "Error checking transactions");
    return false;
  }

  DBG_INFO(0, "Importing transactions...");
  if (!imp.importContext()) {
    DBG_INFO(0, "Error importing transactions");
    return false;
  }

  _flagStaff->accountsUpdated();
  _flagStaff->payeesUpdated();

  return true;
}



void KBanking::invokeHelp(const char *subject){
}



void KBanking::infoTransactionsChanged(Account *a){
  _flagStaff->accountsUpdated();
  App::infoTransactionsChanged(a);
}



void KBanking::outboxCountChanged(int count){
  _flagStaff->outboxCountChanged(count);
}



void KBanking::statusMessage(const QString &s){
  _flagStaff->statusMessage(s);
}



void KBanking::rebuildExportMenu(){
  _flagStaff->rebuildExportMenu();
}



int KBanking::print(const char *docTitle,
                    const char *docType,
                    const char *descr,
                    const char *text){
  GWEN_BUFFER *buf1;
  GWEN_BUFFER *buf2;
  int rv;

  buf1=GWEN_Buffer_new(0, strlen(descr)+32, 0, 1);
  if (!_extractHTML(descr, buf1)) {
    descr=GWEN_Buffer_GetStart(buf1);
  }
  buf2=GWEN_Buffer_new(0, strlen(text)+32, 0, 1);
  if (!_extractHTML(text, buf2)) {
    text=GWEN_Buffer_GetStart(buf2);
  }

  PrintDialog pdlg(this, docTitle, docType, descr, text, 0,
                   "printdialog", true);

  if (pdlg.exec()==QDialog::Accepted)
    rv=0;
  else
    rv=AB_ERROR_USER_ABORT;

  GWEN_Buffer_free(buf2);
  GWEN_Buffer_free(buf1);
  return rv;
}



std::string KBanking::QStringToUtf8String(const QString &qs) {
  std::string result;
  QCString utfData=qs.utf8();
  int len;
  int i;

  len=utfData.length();
  for (i=0; i<len; i++)
    result+=(unsigned char)utfData[i];

  return result;
}



int KBanking::init(){
  GWEN_DB_NODE *db;
  int rv;

  rv=App::init();
  if (rv)
    return rv;

  if (getLastVersion()<
      ((0<<24)+
       (9<<16)+
       (18<<8)+
       5)) {
    DBG_NOTICE(0, "Updating from old version, updating transactions");
    CheckDuplicates w(this);
    w.show();
    w.adjustDates();
  }

  db=getAppData();
  assert(db);
  db=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT,
                      "kbanking");
  assert(db);
  _lastReportId=GWEN_DB_GetIntValue(db, "lastReportId", 0, 0);

  _sampleReportModules();

  return 0;
}



int KBanking::fini(){
  GWEN_DB_NODE *db;

  db=getAppData();
  assert(db);
  db=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT,
                      "kbanking");
  assert(db);
  GWEN_DB_SetIntValue(db, GWEN_DB_FLAGS_OVERWRITE_VARS,
                      "lastReportId",
                      _lastReportId);

  return App::fini();
}



void KBanking::setMainWindow(MainWindow *w){
  _mainWindow=w;
}



MainWindow *KBanking::getMainWindow(){
  return _mainWindow;
}



Report *KBanking::findReport(const char *name){
  std::list<Report*>::iterator it;

  for (it=_reports.begin(); it!=_reports.end(); it++) {
    std::string s;

    s=QStringToUtf8String((*it)->name());
    if (strcasecmp(s.c_str(), name)==0)
      return *it;
  } // for

  return 0;
}



std::list<Report*> KBanking::getReports(){
  return _reports;
}



GWEN_DB_NODE *KBanking::getReportProfiles(Report *r) {
  GWEN_DB_NODE *db;
  GWEN_BUFFER *nbuf;
  std::string name;

  name=QStringToUtf8String(r->name());
  db=getAppData();
  assert(db);
  db=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT,
                      "kbanking/reports");
  assert(db);
  nbuf=GWEN_Buffer_new(0, 256, 0, 1);
  if (GWEN_Text_EscapeToBuffer(name.c_str(), nbuf)) {
    DBG_ERROR(0, "Could not escape report name");
    GWEN_Buffer_free(nbuf);
    return 0;
  }
  GWEN_Buffer_AppendString(nbuf, "/profiles");
  db=GWEN_DB_GetGroup(db, GWEN_DB_FLAGS_DEFAULT,
                      GWEN_Buffer_GetStart(nbuf));
  assert(db);
  GWEN_Buffer_free(nbuf);

  return db;
}



GWEN_DB_NODE *KBanking::findReportProfile(Report *r, const char *id){
  GWEN_DB_NODE *db;

  db=getReportProfiles(r);
  if (!db)
    return 0;
  db=GWEN_DB_FindFirstGroup(db, "profile");
  while(db) {
    const char *s;

    s=GWEN_DB_GetCharValue(db, "id", 0, 0);
    if (!s) {
      DBG_WARN(0, "Report with no id detected");
    }
    else {
      if (strcasecmp(s, id)==0)
        return db;
    }
    db=GWEN_DB_FindNextGroup(db, "profile");
  } // while

  return 0;
}



int KBanking::addReportProfile(Report *r, GWEN_DB_NODE *dbProfile){
  GWEN_DB_NODE *db;
  char numbuf[32];

  db=getReportProfiles(r);
  if (!db)
    return -1;

  snprintf(numbuf, sizeof(numbuf), "%d", ++_lastReportId);
  GWEN_DB_SetCharValue(dbProfile, GWEN_DB_FLAGS_OVERWRITE_VARS,
                       "id", numbuf);
  if (strcasecmp(GWEN_DB_GroupName(dbProfile), "profile")!=0)
    GWEN_DB_GroupRename(dbProfile, "profile");
  GWEN_DB_AddGroup(db, dbProfile);

  return 0;
}



void KBanking::removeReportProfile(Report *r,
                                   GWEN_DB_NODE *dbProfile){
  GWEN_DB_UnlinkGroup(dbProfile);
  GWEN_DB_Group_free(dbProfile);
}



void KBanking::_sampleReportModules(){
  Report *r;

  r=new SimpleReport(this);
  _reports.push_back(r);
}



void KBanking::addTransactionWindow(TransactionWindow *w) {
  _tWindows.push_back(w);
}



void KBanking::delTransactionWindow(TransactionWindow *w) {
  std::list<RefPointer<TransactionWindow> >::iterator it;

  for (it=_tWindows.begin();
       it!=_tWindows.end();
       it++) {
    if ((*it).ptr()==w) {
      // found
      _tWindows.erase(it);
      return;
    }
  }
}



void KBanking::delTransactionWindows() {
  std::list<RefPointer<TransactionWindow> >::iterator it;

  while(!_tWindows.empty()) {
    RefPointer<TransactionWindow> wp=_tWindows.front();
    wp.ref().close();
  }
}














